/*
 * Copyright (C) 2014 Mikel Penagarikano
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package edu.gtts.mediaevalquesst2015;

import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.Map;
import java.util.Set;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;


/**
 *
 * @author mpenagar
 */
public class STDSystem {
	public String name;
	public int[] modelIndex;
	public int[] testIndex;
	public double[] score;
	public boolean[] isTarget;
	public boolean[] decision;
	private int size;
	private static final int initialCapacity = 100;
	public double minScore = Double.POSITIVE_INFINITY;
	public double maxScore = Double.NEGATIVE_INFINITY;
	
	private STDSystem(){
		name = "no name";
		size = 0;
		score = new double[initialCapacity];
		modelIndex = new int[initialCapacity];
		testIndex = new int[initialCapacity];
		isTarget = new boolean[initialCapacity];
		decision = new boolean[initialCapacity];
	}
	
	private void addScore(int modelIndex, int testIndex, double score, boolean isTarget, boolean decision){
		if (size == this.score.length) {
			int newLength = ((size+1) * 4) / 3;
			this.score = Arrays.copyOf(this.score,newLength);
			this.modelIndex = Arrays.copyOf(this.modelIndex,newLength);
			this.testIndex = Arrays.copyOf(this.testIndex,newLength);
			this.isTarget = Arrays.copyOf(this.isTarget,newLength);
			this.decision = Arrays.copyOf(this.decision,newLength);
		}
		this.score[size] = score;
		this.modelIndex[size] = modelIndex;
		this.testIndex[size] = testIndex;
		this.isTarget[size] = isTarget;
		this.decision[size++] = decision;
		
		if (score > maxScore) maxScore = score;
		else if (score < minScore) minScore = score;
	}

	private void shrinkToSize(){
		if (size < score.length) {
			score = Arrays.copyOf(score, size);
			modelIndex = Arrays.copyOf(modelIndex, size);
			testIndex = Arrays.copyOf(testIndex, size);
			isTarget = Arrays.copyOf(isTarget, size);
			decision = Arrays.copyOf(decision, size);
		}
	}
	
	public static STDSystem load(String fileName,Map<String,Integer> modelMap,Set<String> nullModelSet,Map<String,Integer> testMap,boolean[][] groundTruth,Double defaultScore,boolean quiet) throws ParserConfigurationException, SAXException, IOException{
		SAXParserFactory fac = SAXParserFactory.newInstance();
		fac.setValidating(false);
		SAXParser parser = fac.newSAXParser();
		StdlistHandler handler = new StdlistHandler(modelMap,nullModelSet,testMap,groundTruth,defaultScore);
		try {
			if (!quiet) System.out.print("Loading stdlist file: " + fileName + " ... ");
			parser.parse(new File(fileName),handler);
			
			if (!quiet) System.out.println("Done");
		} catch (SAXParseException ex) {
			System.err.println("Parse Exception at line " + ex.getLineNumber() + " while parsing \"" + fileName + "\" :\n\n " + ex.getLocalizedMessage() + "\n");
			System.exit(1);
		}
		return handler.getSystem(quiet);
	}

	
	public TermWeightedValue estimateTWV(double beta){
		return new TermWeightedValue(score,modelIndex,isTarget,decision,beta);
	}
	
	
	private static class StdlistHandler extends LocatedHandler {
		private final STDSystem sys;
		private final Map<String,Integer> modelMap;
		private final Map<String,Integer> testMap;
		private final Set<String> nullModelSet;
		private final Double defaultScore;
		private final boolean[][] groundTruth;
		private final boolean[][] done;
		private Integer modelIndex;
		private boolean nullModel = false;
		
		public StdlistHandler(Map<String,Integer> modelMap, Set<String> nullModelSet, Map<String,Integer> testMap, boolean[][] groundTruth, Double defaultScore){
			this.modelMap = modelMap;
			this.nullModelSet = nullModelSet;
			this.testMap = testMap;
			this.groundTruth = groundTruth;
			this.defaultScore = defaultScore;
			sys = new STDSystem();
			done = new boolean[modelMap.size()][testMap.size()];
		}
		
		
		public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {
			if ((!nullModel) && qName.equals("term")) {

				String s = attributes.getValue("file");
				checkAttribute("file", s);
				Integer testIndex = testMap.get(s);
				if (testIndex == null) {
					throwLocatedSAXParseException("unknown test name: \"" + s + "\"");
				}

				s = attributes.getValue("decision");
				checkAttribute("decision", s);
				boolean decision = false;
				if (s.equals("YES")) decision = true;
				else if (s.equals("NO")) decision = false;
				else throwLocatedSAXParseException("Wrong value for attribute \"decision\" : " + s);

				if (done[modelIndex][testIndex]) {
					throwLocatedSAXParseException("Duplicated trial");
				}

				s = attributes.getValue("score");
				checkAttribute("score", s);
				double score = 0.0;
				try {
					score = Double.parseDouble(s);
				} catch (NumberFormatException e) {
					throwLocatedSAXParseException("Wrong number format for attribute \"score\" : " + s);
				}				
				
				sys.addScore(modelIndex,testIndex,score, groundTruth[modelIndex][testIndex], decision);				
				done[modelIndex][testIndex] = true;
				
			} else if (qName.equals("detected_termlist")) {
				String s = attributes.getValue("termid");
				checkAttribute("termid", s);
				if (nullModelSet.contains(s)) nullModel = true;
				else {
					nullModel = false;
					modelIndex = modelMap.get(s);
					if (modelIndex == null) throwLocatedSAXParseException("unknown model name: \"" + s + "\"");
				}
			} else if (qName.equals("stdlist")) {
				String s = attributes.getValue("system_id");
				sys.name = s;
			}
		}
		
		public STDSystem getSystem(boolean quiet){
			int n = sys.size;
			double minScore = sys.minScore;
			double maxScore = sys.maxScore;
			String[] model = new String[modelMap.size()];
			String[] test = new String[testMap.size()];
			for (Map.Entry<String,Integer> e : modelMap.entrySet()) model[e.getValue()] = e.getKey();
			for (Map.Entry<String,Integer> e : testMap.entrySet()) test[e.getValue()] = e.getKey();

			for (int mInd = 0; mInd < model.length; mInd++)
				for (int tInd = 0; tInd < test.length; tInd++)
					if (!done[mInd][tInd]) {
						if (defaultScore == null) {
							System.err.println("Score not found for model: \"" + model[mInd] + "\" and test: \"" + test[tInd] + "\"\nNo default score, EXITING!");
							System.exit(1);
						} else sys.addScore(mInd,tInd,defaultScore, groundTruth[mInd][tInd], false);
					}
			if (!quiet) {
				int N = sys.size;
				System.out.println("Scores: " + N + " (" + model.length + " model x " + test.length + " test)  given: " + n + "  default(" + defaultScore + "): " + (N-n) + "  minScore: " + minScore + "  maxScore: " + maxScore);
			}
			sys.shrinkToSize();
			return sys;
		}
	}

	/**
	 * Split STD System by model
	 * @return an array of STD systems, one per model. All the model insdexes
	 * in the STD systems are set to 0
	 */
	public STDSystem[] split(){
		int nModel = 0;
		for (int i : modelIndex) if (i>nModel) nModel = i; nModel++;
		STDSystem[] sv = new STDSystem[nModel];
		for (int i = 0; i < sv.length; i++) sv[i] = new STDSystem();
		for (int i = 0; i < size; i++)
			//sv[modelIndex[i]].addScore(modelIndex[i], testIndex[i], score[i], isTarget[i], decision[i]);
			sv[modelIndex[i]].addScore(0, testIndex[i], score[i], isTarget[i], decision[i]);
		for (STDSystem s : sv) s.shrinkToSize();
		return sv;
	}
	
	/**
	 * Joins in a single System a set of model dependent STD Systems. Model id-s
	 * are set to the corresponding array index
	 * @param ss STD System array to split
	 * @return The resulting STD System
	 */
	public static STDSystem join(STDSystem[] ss){
		STDSystem s = new STDSystem();
		for (int i = 0; i < ss.length; i++){
			STDSystem x = ss[i];
			for (int j = 0; j < x.size; j++)
				//s.addScore(x.modelIndex[i], x.testIndex[i], x.score[i], x.isTarget[i], x.decision[i]);
				s.addScore(i, x.testIndex[j], x.score[j], x.isTarget[j], x.decision[j]);
		}
		s.shrinkToSize();
		return s;
	}
	
}